# run with : python3 -m unittest TestEvidence

import unittest
import os
import numpy as np
import math
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt
from BayesicFitting import *


class TestEvidence( unittest.TestCase  ) :

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testEvidence( self ) :
        print( "====testEvidence======================" )
        nn = 100
        x = np.arange( nn, dtype=float ) / (nn/2) - 1
        ym = 1.2 + 0.5 * x
        nf = 0.1
        np.random.seed( 2345 )
        noise = np.random.randn( nn )

        y = ym + nf * noise

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )
        w = np.ones( nn, dtype=float )

        pars = bf.fit( y, w )
        print( "pars  ", pars )
        yfit = pm.result( x, pars )
        print( "stdv  ", bf.getStandardDeviations() )
        bf.getHessian()
        bf.chiSquared( y, weights=w )
        print( "chisq %f  scale %f  sumwgt %f" % ( bf.chisq, bf.getScale(), bf.sumwgt ) )

        lolim = [-100.0,-100.0]
        hilim = [+100.0,+100.0]
        nslim = [0.01, 100.0]

        print( "=== Evidence for Parameters only; fixedScale is not set" )
        print( "evid %f  occam %f  lhood %f  fixedScale="%
            ( bf.getLogZ(  limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood ), bf.fixedScale )

        if self.doplot :
            plt.plot( x, ym, 'g.' )
            plt.plot( x, y, 'b+' )
            plt.plot( x, yfit, 'r-' )
            plt.show()


        print( "=== Evidence for Parameters and scale; fixedScale not set" )
        print( "evid %f  occam %f  lhood %f  fixedScale="%
            ( bf.getLogZ(  limits=[lolim,hilim], noiseLimits=nslim ),
            bf.logOccam, bf.logLikelihood ), bf.fixedScale )
#        self.assertRaises( ValueError, bf.getLogZ() )
#        self.assertRaises( ValueError, bf.logOccam )
#        self.assertRaises( ValueError, bf.logLikelihood )

        print( "=== Evidence for Parameters and scale; fixedScale=10" )
        bf.fixedScale = 10.0
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim], noiseLimits=nslim ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        print( "=== Evidence for Parameters, fixed scale = 1" )
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        print( "=== Evidence for Parameters and fixed scale = 0.01" )
        bf.fixedScale = 0.01
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        print( "=== Evidence for Parameters and fixed scale = 0.1" )
        bf.fixedScale = 0.1
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        print( "=== Evidence for Parameters and fixed scale = 1.0" )
        bf.fixedScale = 1.0
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        print( "=== Evidence for Parameters and fixed scale = 10.0" )
        bf.fixedScale = 10.0
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( limits=[lolim,hilim] ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )

        pm.setLimits( lolim, hilim )
        print( "limits par[0]  ", pm.priors[0].lowLimit, pm.priors[0].highLimit )
        print( "limits par[1]  ", pm.priors[1].lowLimit, pm.priors[1].highLimit )

        print( "=== Evidence for Parameters and fixed scale = 10.0" )
        bf.fixedScale = 10.0
        print( "evid %f  occam %f  lhood %f  fixeScale=%f" %
            ( bf.getLogZ( ),
            bf.logOccam, bf.logLikelihood, bf.fixedScale ) )


    def testSimple1( self ) :
        print( "====testEvidence for Gauss (Simple) ====" )
        np.random.seed( 2345 )
        nn = 100
        x = np.arange( nn, dtype=int ) // 20
#        ym = np.linspace( 1.0, 2.0, nn )
        ym = 1.5
        nf = 0.1
        noise = np.random.normal( 0.0, 1.0, nn )

        y = ym + nf * noise

        pm = PolynomialModel( 0 )
        print( pm.npchain )
        bf = Fitter( x, pm, fixedScale=nf )

        pars = bf.fit( y )
        print( "pars  ", pars )
        yfit = pm.result( x, pars )
        std = bf.stdevs
        print( "stdv  ", std )
        print( "chisq %f  scale %f  sumwgt %f" % ( bf.chisq, bf.scale, bf.sumwgt ) )
        print( bf.chiSquared( y, pars ) )

        lo = 1.40
        hi = 1.60
        logz = bf.getLogZ( limits=[lo,hi] )
        maxloglik = bf.logLikelihood
        lintpr = math.log( hi - lo )

        errdis = GaussErrorDistribution( scale=nf )
        problem = ClassicProblem( model=pm, xdata=x, ydata=y )

        npt = 401
        p0 = np.linspace( lo, hi, npt )

        for i in range( pm.npchain ) :
            L0 = np.ndarray( npt, dtype=float )
            pp = np.append( pars, [nf] )
            for k,p in enumerate( p0 ) :
                pp[i] = p
                L0[k] = errdis.logLikelihood( problem, pp )
            lz = np.log( np.sum( np.exp( L0 ) ) * (hi - lo) / npt )
            maxl = np.max( L0 )
            L0 -= maxloglik

            if self.doplot :
                gm = GaussModel()
                gm.parameters = [1.0, pars[i], std[i]]
                plt.plot( p0, gm( p0 ), 'b-' )
                plt.plot( p0, np.exp( L0 ), 'k-' )

        print( "BF logL  ", maxloglik, bf.logOccam, maxl, np.exp( maxl ), lintpr )
        print( "ED logL  ", errdis.logLikelihood( problem, np.append( pars, [nf] ) ) )
        print( "evid      %f  %f  %f"%( logz, lz - lintpr,
                maxloglik + math.log( 0.01 * math.pi ) - lintpr  ) )
#        print( "evid      %f  %f  %f"%( logz, lz, maxloglik + math.log( 0.01 * math.pi )  ) )

        if self.doplot :
            plt.show()


    def testSimpleLaplace1( self ) :
        print( "====testEvidence for Laplace (Simple 1) ====" )
        np.random.seed( 2345 )
        nn = 1000
        x = np.arange( nn, dtype=int ) // 20
        ym = np.linspace( 1.0, 2.0, nn )
        nf = 0.1
        noise = np.random.laplace( 0.0, 1.0, nn )

        y = ym + nf * noise

        pm = PolynomialModel( 0 )
        print( pm.npchain )
        bf = PowellFitter( x, pm, scale=nf, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", pars )
        yfit = pm.result( x, pars )
        std = bf.stdevs
        print( "stdv  ", std )
        print( np.median( y ), np.mean( y ) )
        print( "scale ", bf.fixedScale  )

        print( "scale %f  sumwgt %f" % ( bf.scale, bf.sumwgt ) )


        lo = 1.45
        hi = 1.55
        logz = bf.getLogZ( limits=[lo,hi] )
        maxloglik = bf.logLikelihood
        lintpr = math.log( hi - lo )

        errdis = LaplaceErrorDistribution( scale=nf )
        problem = ClassicProblem( model=pm, xdata=x, ydata=y )

        npt = 101
        p0 = np.linspace( lo, hi, npt )

        for i in range( pm.npchain ) :
            L0 = np.ndarray( npt, dtype=float )
            pp = np.append( pars, [nf] )
            for k,p in enumerate( p0 ) :
                pp[i] = p
                L0[k] = errdis.logLikelihood( problem, pp )
            lz = np.log( np.sum( np.exp( L0 ) ) * (hi - lo) / npt )
            maxl = np.max( L0 )
            L0 -= maxl

            if self.doplot :
                gm = GaussModel()
                gm.parameters = [1.0, pars[i], std[i]]
#                gm.parameters = [1.0, pars[i], 2*nf/math.sqrt(nn) ]
                plt.plot( p0, gm( p0 ), 'b-' )
                plt.plot( p0, np.exp( L0 ), 'k-' )

        print( "BF logL  ", maxloglik, bf.logOccam, maxl, np.exp( maxl ), lintpr )
        print( "ED logL  ", errdis.logLikelihood( problem, np.append( pars, [nf] ) ) )
        print( "evid      %f  %f  %f"%( logz, lz - lintpr,
                maxloglik + math.log( 0.01 * math.pi ) - lintpr  ) )
#        print( "evid      %f  %f  %f"%( logz, lz, maxloglik + math.log( 0.01 * math.pi )  ) )

        if self.doplot :
            plt.show()


    def testGauss( self ) :
        print( "====testEvidence for Gauss============" )
        nn = 1000
        x = np.arange( nn, dtype=int ) // 20
        ym = np.linspace( 1.0, 2.0, nn )
        nf = 0.1

        np.random.seed( 2345 )
        noise = np.random.normal( 0.0, 1.0, nn )

        y = ym + nf * noise

#        pm = FreeShapeModel( nn // 20 )
        pm = PolynomialModel( 1 )
        bf = AmoebaFitter( x, pm, errdis="gauss" )

        pars = bf.fit( y )
        print( "pars  ", pars )
        yfit = pm.result( x, pars )
        std = bf.stdevs
        print( "stdv  ", std )
        print( "scale %f  sumwgt %f" % ( bf.scale, bf.sumwgt ) )

        errdis = GaussErrorDistribution( scale=nf )
        problem = ClassicProblem( model=pm, xdata=x, ydata=y )

        p0 = np.linspace( 0.0, 3.0, 301 )
        for i in range( pm.npchain ) :
            L0 = np.ndarray( 301, dtype=float )
            pp = np.append( pars, [0.1] )
            for k,p in enumerate( p0 ) :
                pp[i] = p
                L0[k] = errdis.logLikelihood( problem, pp )
            L0 -= np.max( L0 )

            if self.doplot :
                gm = GaussModel()
                gm.parameters = [1.0, pars[i], std[i]]
                plt.plot( p0, gm( p0 ), 'b-' )
                plt.plot( p0, np.exp( L0 ), 'k-' )

        if self.doplot :
            plt.show()


    def testLaplace( self ) :
        print( "====testEvidence for Laplace============" )
        nn = 100
        x = np.arange( nn, dtype=float ) / (nn/2) - 1
        ym = 1.3
        nf = 0.1
        np.random.seed( 2345 )
        noise = np.random.laplace( 0.0, 1.0, nn )

        y = ym + nf * noise

        pm = PolynomialModel( 0 )
        bf = PowellFitter( x, pm, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-10 )
        print( "pars  ", pars )
        yfit = pm.result( x, pars )
        std = bf.stdevs
        print( "stdv  ", std )
        print( "scale %f  sumwgt %f" % ( bf.scale, bf.sumwgt ) )

        errdis = LaplaceErrorDistribution( scale=nf )
        problem = ClassicProblem( model=pm, xdata=x, ydata=y )

        p0 = np.linspace( 1.2, 1.5, 201 )
        L0 = np.ndarray( 201, dtype=float )
        for k,p in enumerate( p0 ) :
            pp = [p, 0.1]
            L0[k] = errdis.logLikelihood( problem, pp )
        L0 -= np.max( L0 )

        if self.doplot :
            gm = GaussModel()
            gm.parameters = [1.0, pars[0], std[0]]
            plt.plot( p0, gm( p0 ), 'b-' )
            plt.plot( p0, np.exp( L0 ), 'k-' )

            plt.show()


if __name__ == '__main__':
    unittest.main( )




# run with : python3 -m unittest TestBaseModel.py

import unittest
import numpy as numpy
from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

class TestBaseModel( unittest.TestCase ) :

    def test_init( self ) :
        bm = BaseModel( 5 )
        self.assertTrue( bm.npbase == 5 )
        self.assertTrue( bm.ndim == 1 )
        self.assertTrue( bm.deltaP[0] == 0.00001 )
        self.assertEqual( bm.deltaP.shape, (1,) )
        self.assertTrue( isinstance( bm.deltaP, numpy.ndarray ) )

        self.assertEqual( bm.posIndex.shape, (0,) )
        self.assertTrue( isinstance( bm.posIndex, numpy.ndarray ) )

        self.assertEqual( bm.nonZero.shape, (0,) )
        self.assertTrue( isinstance( bm.nonZero, numpy.ndarray ) )

        bm.posIndex = [2,4]
        bm.nonZero = 3
        self.assertEqual( bm.posIndex.shape, (2,) )
        self.assertTrue( isinstance( bm.posIndex, numpy.ndarray ) )
        self.assertEqual( bm.posIndex[0], 2 )
        self.assertEqual( bm.posIndex[1], 4 )
        self.assertEqual( bm.nonZero.shape, (1,) )
        self.assertTrue( isinstance( bm.nonZero, numpy.ndarray ) )
        self.assertEqual( bm.nonZero[0], 3 )

        cm = BaseModel( copy=bm )

        self.assertTrue( cm.npbase == 5 )
        self.assertTrue( cm.ndim == 1 )
        self.assertTrue( cm.deltaP[0] == 0.00001 )
        self.assertEqual( cm.deltaP.shape, (1,) )
        self.assertTrue( isinstance( cm.deltaP, numpy.ndarray ) )

        self.assertEqual( cm.posIndex.shape, (2,) )
        self.assertTrue( isinstance( cm.posIndex, numpy.ndarray ) )
        self.assertEqual( cm.posIndex[0], 2 )
        self.assertEqual( cm.posIndex[1], 4 )
        self.assertEqual( cm.nonZero.shape, (1,) )
        self.assertTrue( isinstance( cm.nonZero, numpy.ndarray ) )
        self.assertEqual( cm.nonZero[0], 3 )

    def test_parnames( self ) :
        bm = GaussModel( )
#        pm = PolynomialModel( 1 )

        Tools.printclass( bm )
#        Tools.printclass( pm )

#        bm += pm

        self.assertTrue( bm.amplitude == bm.parameters[0] )
        self.assertTrue( bm.center == bm.parameters[1] )
        self.assertTrue( bm.width == bm.parameters[2] )

#        self.assertTrue( bm.parameter_0 == bm.parameters[3] )
#        self.assertTrue( bm.parameter_1 == bm.parameters[4] )




if __name__ == '__main__':
    unittest.main()



import numpy as numpy
import math
import warnings

from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2020 Do Kester

class Kepplers2ndLaw( object ):
    """
    Class for calculating Kepplers second law for planetary motion.

    The projection of the orbit on the sky is not included in this class.
.
    The algorithm was taken from
        Cory Boule etal. (2017) J. of Double Star Observations Vol 13 p.189.

        http://www.jdso.org/volume13/number2/Harfenist_189-199.pdf

    p_0 : e     eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
    p_1 : a     semi major axis (>0)
    p_2 : P     period of the orbit (>0)
    p_3 : T     phase since periastron passage (0<p_3<2 pi)

    The parameters are initialized at [0.0, 1.0, 1.0, 0.0].

    """
    TWOPI = 2 * math.pi
    MAXITER = 100

    def meanAnomaly( self, xdata, params ) :
        """
        Return the mean anomaly.

        P = params[2] = period
        p = params[3] = periastron passage

        M = 2 * pi * xdata / P - p

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, period, ppass
        """
        period = params[2]
        phase = params[3]

        ## return mean anomaly
        return self.TWOPI * xdata / period - phase

    def dMdx( self, xdata, params ) :
        """
        Return derivatives of M (mean anomaly) to xdata

        Returns
        -------
        dMdx : array_like
            derivatives of M to x (xdata)
        """
        period = params[2]
        return numpy.zeros_like( xdata ) + self.TWOPI / period


    def dMdpar( self, xdata, params ) :
        """
        Return derivatives of M (mean anomaly) to relevant parameters.

        Returns
        -------
        dMdP : array_like
            derivatives of M to P (period)
        dMdp : array_like
            derivatives of M to p (phase of periastron)
        """
        period = params[2]

        dMdP = - self.TWOPI * xdata / ( period * period )
        dMdp = -1
        return ( dMdP, dMdp )


    def eccentricAnomaly( self, xdata, params, Estart=None ) :
        """
        Take the best one : Halleys method

        It converges in a few iterations for e <= 0.999999999
        """
        return self.eccentricAnomaly2( xdata, params, Estart=Estart )


    def eccentricAnomaly0( self, xdata, params ) :
        """
        Return the eccentric anomaly, i.e. the solution for E of

        Standard method by Jean Meuss :
            Astronomical Algorithms, 2nd ed.,
            Willmann-Bell, Inc, Virginia, 193-196, 397-399

        e = params[0] = eccentricity
        M = mean anomaly

        E = M + e * sin( E )

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, period, ppass
        """
        M = self.meanAnomaly( xdata, params )
        eccen = params[0]

        Ep = M[:]
        iter = 0
        ## calculate the eccentric anomaly E
        while iter < self.MAXITER :
            sinE = numpy.sin( Ep )
            E = M - eccen * sinE
            iter += 1
            if all( numpy.abs( E - Ep ) < 1e-8 ) : break
            Ep = E
        else :
            print( "EA0: No convergence at iter %d for eccentricity %8.4f" % ( iter, eccen ) )

        setatt( self, "iter", iter )
        return E

    def eccentricAnomaly1( self, xdata, params ) :
        """
        Newtons method.

        Return the eccentric anomaly, i.e. the solution for E of

        e = params[0] = eccentricity
        M = mean anomaly

        E = M + e * sin( E )

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, period, ppass
        """
        M = self.meanAnomaly( xdata, params )
        eccen = params[0]

        Ep = M[:]
        iter = 0
        ## calculate the eccentric anomaly E
        while iter < self.MAXITER :
            sinE = numpy.sin( Ep )
            cosE = numpy.cos( Ep )
            E = Ep - ( M + eccen * sinE - Ep ) / ( eccen * cosE - 1 )
            iter += 1
            if all( numpy.abs( E - Ep ) < 1e-8 ) : break
            Ep = E
        else :
            print( "EA1: No convergence at iter %d for eccentricity %8.4f" % ( iter, eccen ) )

        setatt( self, "iter", iter )
        return E

    def eccentricAnomaly2( self, xdata, params, Estart=None ) :
        """
        Halleys method.

        Return the eccentric anomaly, i.e. the solution for E of

        e = params[0] = eccentricity
        M = mean anomaly

        E = M + e * sin( E )

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, period, phase
        Estart : array_like
            starting values for E
        """
        M = self.meanAnomaly( xdata, params )
        eccen = params[0]


        E = M[:] if Estart is None else Estart

        iter = 0
        ## calculate the eccentric anomaly E
        while iter < self.MAXITER :
            Ep = E
            sinE = numpy.sin( Ep )
            cosE = numpy.cos( Ep )
            es = eccen * sinE
            fx = M + es - Ep
            fp = eccen * cosE - 1

            E = Ep - 2 * fx * fp / ( 2 * fp * fp + fx * es )
            iter += 1
            if all( numpy.abs( E - Ep ) < 1e-8 ) : break
        else :
#            # uncomment for more diagnostics
#            q = numpy.where( numpy.abs( E - Ep ) > 1e-8 )
#            print( "\n" )
#            print( "params  ", fmt( params, max=None ) )
#            print( "Ep      ", fmt( Ep[q] ) )
#            print( "E       ", fmt( E[q] ) )
#            print( "sinE    ", fmt( sinE[q] ) )
#            print( "cosE    ", fmt( cosE[q] ) )

            print( "EA2: No convergence at iter %d for eccentricity %8.4f" % ( iter, eccen ) )

        setatt( self, "iter", iter )
        setatt( self, "sinE", sinE )
        setatt( self, "cosE", cosE )

        return E

    def dEdM( self, xdata, params, cosE ) :
        """
        Return derivatives of E (eccentric anomaly) to mean anomaly

        Returns
        -------
        dEdM : array_like
            derivatives of E to M (mean anomaly)
        """
        eccen = params[0]
        dE = 1.0 / ( 1.0 - eccen * cosE )
        return dE

    def dEdx( self, xdata, params, cosE ) :
        """
        Return derivatives of E (eccentric anomaly) to xdata

        Returns
        -------
        dEdx : array_like
            derivatives of E to x (xdata)
        """
        return self.dEdM( xdata, params, cosE ) * self.dMdx( xdata, params )

    def dEdpar( self, xdata, params, cosE, sinE ) :
        """
        Return derivatives of E (eccentric anomaly) to relevant parameters.

        Returns
        -------
        dEde : array_like
            derivatives of E to e (eccentricity)
        dEdP : array_like
            derivatives of E to P (period)
        dEdp : array_like
            derivatives of E to p (phase of periastron)
        """
        dEdM = self.dEdM( xdata, params, cosE )
        dMdP, dMdp = self.dMdpar( xdata, params )
        dEdp = dEdM * dMdp
        dEdP = dEdM * dMdP
        dEde = dEdM * sinE
        return ( dEde, dEdP, dEdp )


    def radiusAndTrueAnomaly( self, xdata, params ) :
        """
        Return the radius and the true anomaly.

        e = params[0] = eccentricity
        a = params[1] = semimajor axis
        E = eccentric anomaly

        r = a * ( 1 - e * cos( E ) )

        v = 2 * arctan( sqrt( (1+e)/(1-e) ) * tan( E / 2 ) )

        from Wikepedia => Trigoniometic Identities :
        tan( E / 2 ) = sqrt( ( 1 - cos( E ) ) / ( 1 + cos( E ) ) )
                     = sqrt( ( 1 - c ) * ( 1 + c ) / ( 1 + c )^2 )
                     = sqrt( s^2 / ( 1 + c )^2 )
                     = s / ( 1 + c )
                     = sin( E ) / ( 1 + cos( E ) )
        Avoid cases where cos( E ) is too close to -1

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, inclin, ascpos, asclon, period, ppass

        Returns
        -------
        r : array_like
            radius
        v : array_like
            true anomaly

        """
        E = self.eccentricAnomaly( xdata, params )

        eccen = params[0]
        semimaj = params[1]

        ## r = radius
        cosE = self.cosE
        r = semimaj * ( 1 - eccen * cosE )

        ## v = true anomaly
        ef = math.sqrt( ( 1 + eccen ) / ( 1 - eccen ) )

        v = 2 * numpy.arctan2( ef * self.sinE, 1 + cosE )

        setatt( self, "eccAnomaly", E )
        ## return radius and true anomaly
        return ( r, v )

    def drvdE( self, xdata, params, cosE, sinE ) :
        """
        Return derivatives of r (radius) and v (true anomaly) to eccentric anomaly

        Parameters
        ----------
        xdata : array_like
            times in the orbit
        params : array_like
            parameters: eccentr, semimajor, period, ppass
        cosE : array_like
            cosine of E
        sinE : array_like
            sine of E

        Returns
        -------
        drdE : array_like
            derivatives of r to E (eccentric anomaly)
        dvdE : array_like
            derivatives of v to E (eccentric anomaly)
        """

        eccen = params[0]
        semimaj = params[1]

        dr = semimaj * eccen * sinE

        ef = ( 1 + eccen ) / ( 1 - eccen )
        c2 = numpy.where( cosE <= -0.9999999999, 1e10, 1 / ( 1 + cosE ) )
        t2 = ( 1 - cosE ) * c2          ## = tan^2( E/2 )       from Wiki:
        c2 *= 2                         ## = 1 / cos^2( E/2 )   trig identities
        dv = math.sqrt ( ef ) * c2
        dv /= ( ef * t2 + 1 )

        return ( dr, dv )

    def drvdx( self, xdata, params, cosE, sinE ) :
        """
        Return derivatives of r (radius) and v (true anomaly) to xdata

        Returns
        -------
        drdx : array_like
            derivatives of r to x (xdata)
        dvdx : array_like
            derivatives of v to x (xdata)
        """

        drvdE = self.drvdE( xdata, params, cosE, sinE )
        dEdx = self.dEdx( xdata, params, cosE )
        return ( drvdE[0] * dEdx, drvdE[1] * dEdx )

    def drvdpar( self, xdata, params, E, cosE, sinE ) :
        """
        Return derivatives of r (radius) and v (true anomaly) to relevant parameters.

        Returns
        -------
        drde : array_like
            derivatives of r to e (eccentricity)
        drda : array_like
            derivatives of r to a (semimajor axis)
        drdP : array_like
            derivatives of r to P (period)
        drdp : array_like
            derivatives of r to p (phase of periastron)
        dvde : array_like
            derivatives of v to e (eccentricity)
        dvdP : array_like
            derivatives of v to P (period)
        dvdp : array_like
            derivatives of v to p (phase of periastron)
        """
        eccen = params[0]
        semimaj = params[1]

        ( dEde, dEdP, dEdp ) = self.dEdpar( xdata, params, cosE, sinE )
        ( drdE, dvdE ) = self.drvdE( xdata, params, cosE, sinE )

        # r = semimaj * ( 1 - eccen * cosE )
        drdP = drdE * dEdP
        drdp = drdE * dEdp
        drde = drdE * dEde
        drde -= semimaj * cosE
        drda = 1 - eccen * cosE

        dvdP = dvdE * dEdP
        dvdp = dvdE * dEdp

        ef = ( 1 + eccen ) / ( 1 - eccen )
        sef = math.sqrt( ef )
        e2 = ( 1 - eccen ) * ( 1 - eccen )

        c2 = numpy.where( cosE <= -0.9999999999, 1e10, 1 / ( 1 + cosE ) )

        tanE = sinE * c2                ## for tan( E/2 )

        dd1 = 2 * tanE / ( sef * e2 )
        dd2 = dd1 + 2 * sef * dEde * c2
        dvde = dd2 / ( 1 + ef * tanE * tanE )
        dvde = numpy.where( numpy.isnan( dvde ), 0.0, dvde )

        return ( drde, drda, drdP, drdp, dvde, dvdP, dvdp )


    def getMsini( self, stellarmass ) :
        """
        Return the mass of the exoplanet in Jupiter masses.

        Parameters
        ----------
        stellarmass : float
            mass of the host star in solar masses.
        """
        p = self.parameters
        return ( 4.91e-3 * math.pow( stellarmass, 2.0 / 3.0 ) * p[0] *
                 math.pow( p[1], 1.0 / 3.0 ) * math.sqrt( 1 - p[2] * p[2] ) )






import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt

from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2020 Do Kester


class LogisticModel( NonLinearModel ):
    """
    Logistic Model.

        f( x:p ) = p_0 / ( 1 + exp( ( x - p_1 ) / p_2 ) )

    where

        p_0 : amplitude
        p_1 : center
        p_2 : slope

    The parameters are initialized at {1.0, 0.0, 1.0}.

    Examples
    --------
    >>> lm = LogisticModel( )
    >>> print( lm )
    Logistic: f( x:p ) = p_0 / ( 1 + exp( ( p_1 - x ) / p_2 ) )
    >>> print( lm.npars )
    3
    >>> print( lm( numpy.arange( 11 ) - 5 ) )
    [  3.72665317e-06   3.35462628e-04   1.11089965e-02   1.35335283e-01
       6.06530660e-01   1.00000000e+00   6.06530660e-01   1.35335283e-01
       1.11089965e-02   3.35462628e-04   3.72665317e-06]

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """

    def __init__( self, copy=None, **kwargs ):
        """
        Logistic response model.

        Number of parameters is 3.

        Parameters
        ----------
        copy : LogisticModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        names = ["amplitude", "center", "slope"]
        param = [1.0,0.0,1.0]
        super( LogisticModel, self ).__init__( 3, copy=copy, params=param,
                    names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return LogisticModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = ( params[1] - xdata ) / params[2]
        res = params[0] / ( 1 + numpy.exp( x ) )
        return res

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        partial = numpy.ndarray( ( Tools.length( xdata ), self.npbase ) )
        a = params[0]
        s = params[2]
        x = ( params[1] - xdata )
        e = numpy.exp( x / s )
        f = 1 / ( 1 + e )
        f2 = f * f

        parts = { 0 : ( lambda: f ),
                  1 : ( lambda: -a / s * f2 * e ),
                  2 : ( lambda: a * x / ( s * s ) * f2 * e ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = ( params[1] - xdata ) / params[2]
        e = numpy.exp( x )

        return ( params[0] / params[2] ) * e / numpy.square( 1 + e )

#        return - ( params[0] / params[2] ) * e / numpy.square( 1 + e )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "Logistic: f( x:p ) = p_0 / ( 1 + exp( ( p_1 - x ) / p_2 ) )" )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.yUnit if k == 0 else self.xUnit





import numpy as numpy
import math
from . import Tools
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class ArctanModel( NonLinearModel ):
    """
    Arctangus Model.

        f( x:p ) = p_0 * arctan( p_2 * ( x - p_1 ) )
        p_0 = amplitude;  p_1 = center; p_2 = slope.

    As always x = input.

    The parameters are initialized at {2/pi, 0.0, 1.0}. It is a non-linear model.

    Attributes from Model
    --------------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Example
    -------
    >>> arct = ArctanModel( )
    >>> print( arct.getNumberOfParameters( ) )
    3


    Author:      Do Kester

    """

    def __init__( self, copy=None, **kwargs ):
        """
        Arc-tangus model.

        Number of parameters is 3.

        Parameters
        ----------
        copy : ArctanModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        params = [2/math.pi, 0.0, 1.0]
        names = ["amplitude","center","slope"]
        super( ArctanModel, self ).__init__( nparams=3, copy=copy, params=params,
                                             names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return ArctanModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        result = params[0] * numpy.arctan( params[2] * ( xdata - params[1] ) )
        return result

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.zeros( ( Tools.length( xdata ), np ) )

        x = params[2] * ( xdata - params[1] )
        xf = params[0] / ( 1 + x * x )

        parts = { 0 : ( lambda: numpy.arctan( x ) ),
                  1 : ( lambda: -xf * params[2] ),
                  2 : ( lambda: xf * ( xdata - params[1] ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each input (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        x = params[2] * ( xdata - params[1] )
        return params[0] * params[2] / ( 1 + x * x )

    def baseName( self ):
        """
        Returns a string representation of the model.


        """
        return str( "Arctan: f( x:p ) = p_0 * arctan( p_2 * ( x - p_1 ) )" )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        if k == 2:
            return self.xUnit ** ( -1 )
        return self.xUnit





import numpy as numpy
import scipy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .NoiseScale import NoiseScale

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2022 Do Kester


class CauchyErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Cauchy or Lorentz likelihood.

        f( x ) = s / ( &pi; * ( s^2 + x^2 ) )

    where x = residual and s = scale

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.

        logL = N ( log( s ) - log( &pi; ) ) - sum( log( x^2 + s^2 ) )

    Weights are not possible in this error distribution. They are silently ignored.

    s is a hyperparameter, which might be estimated from the data.

    Author       Do Kester.

    """

    LOGPI = math.log( math.pi )

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, limits=None, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is not fixed.
         copy : CauchyErrorDistribution
            distribution to be copied.

        """

        super( CauchyErrorDistribution, self ).__init__( scale=scale, limits=limits,
            copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return CauchyErrorDistribution( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes. Needed for getScale()

        """
        if name == "res2" or name == "sumweight":
            object.__setattr__( self, name, value )
        else :
            super( CauchyErrorDistribution, self ).__setattr__( name, value )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        False for this distribution.
        """
        return False

    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale as calculated from the residuals.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        res = self.getResiduals( problem, allpars=allpars )
        self.res2 = res * res
        self.sumweight = problem.sumweight

        scale = scipy.optimize.bisect( self.funct, 0.001, 100.0 )
        return scale

    def funct( self, scale ) :
        """
        Internal use, only.
        """
        return ( numpy.sum( numpy.log( self.res2 + scale * scale ) ) -
                self.sumweight * ( 2 * math.log( scale ) + math.sqrt( 2.0 ) ) )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ):
        """
        Return the log( likelihood ) for a Cauchy distribution.
        Cauchy distr : f( x ) = s / ( pi * ( s^2 + x^2 ) )

        where x = residual and s = scale

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem

        """
        self.ncalls += 1

        scale = allpars[-1]
        res = problem.residuals( allpars[:-1] )
        res2 = numpy.square( res )
        return ( problem.ndata * ( math.log( scale ) - self.LOGPI ) -
                 numpy.sum( numpy.log( res2 + scale * scale ) ) )

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        res = problem.residuals( allpars[:-1], mockdata=mockdata )

        scale = allpars[-1]
        s2 = scale * scale
        res2 = res * res
        return math.log( scale ) - self.LOGPI - numpy.log( res2 + s2 )

    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters
        in fitIndex.

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1

        scale = allpars[-1]
        res = problem.residuals( allpars[:-1] )
        r2s = res * res + scale * scale
        dM = problem.partial( allpars[:-1] )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for k in fitIndex :
            if k >= 0 :
                dL[i] = 2 * numpy.sum( res * dM[:,k] / r2s )
                i += 1
            else :
                dL[-1] = problem.ndata / scale - numpy.sum( 2 * scale / r2s )

        return dL

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters
        in fitIndex.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated by the model

        """
        res = problem.residuals( allpars[:-1], mockdata=mockdata )
        scale = allpars[-1]
        r2s = res * res + scale * scale

        dM = problem.partial( allpars[:-1] )
##      TBD import mockdata into partial
#        dM = model.partial( self.xdata, param, mockdata=mockdata )

        for k in fitIndex :
            if k >= 0 :
                yield 2 * res * dM[:,k] / r2s
            else :
                yield 1.0 / scale - 2 * scale / r2s

        return

    def __str__( self ) :
        return "Cauchy error distribution"

